
/*
A component of the cT (TM) programming environment.
(c) Copyright 1989 Carnegie Mellon University.
cT is a trademark of Carnegie Mellon University.
All rights reserved.
May not be copied without the written consent
of Carnegie Mellon University.
*/

/* arithmetic expression analyzer */

#include "compute.h"
#include "commands.h"
#ifndef SKIP
#include "yacc.h"
#endif
#include "ecglobal.h"
#include "cglobals.h"
#include "tglobals.h"
#include "exprdefs.h"


#ifdef ctproto
int  icompile(void);
int  StopProgram(char  *s);
char  FAR *TUTORalloc(long  size,int  abort,char  *label);
int  AllowHandlePurge(unsigned int  mm);
int  TUTORpurge_info(unsigned int  mm,int  ptype,int  (FAR *restoreproc)(),int  restorearg);
char  FAR *GetPtr(unsigned int  mm);
unsigned int  TUTORhandle(char  *name,long  size,int  purgewmrm);
extern void get_cmd_table(void);
extern void get_fct_table(void);
#endif /* ctproto */

#ifdef macproto
extern int icgen68K(void);
extern int icgenPPC(void);
#endif

/* ******************************************************************* */

icompile() /* initialize tables for compile */

{   int i; /* index in expression analyzer table */
    struct exprt FAR *t; /* pointer to current table entry */

    xtokmalloc = 0; /* no token table yet */
    prevend = 0; /* initialize to reasonable values */
    appendok = FALSE;
    userExp = 0; /* not -compute- */

    /* get command and functions tables */

#ifndef EXECUTE
    get_cmd_table();
#endif
    get_fct_table();

    /* allocate operators/operands description table */

    xtH = TUTORhandle("xt",(long)(XT_MAX*sizeof(struct exprt)),TRUE);
    xtP = (struct exprt FAR *) GetPtr(xtH);
    TUTORpurge_info(xtH,M_WORM,FARNULL,0);
    AllowHandlePurge(xtH);
    
    /* allocate command error table */
    /* entries are TRUE if token or command can cause execution error */
    
    cmd_err_tab = (unsigned char FAR *)TUTORalloc((long)C_MAX,TRUE,"cmd_err");
    if (!cmd_err_tab) StopProgram(NEARNULL);
        
    /* set all table entries to defaults */

    for (i=0; i<XT_MAX; i++) {
        t = &xtP[i];
        t->op = TRUE;       /* set for operator */
        t->unary = FALSE;   /* set for binary operator */
        t->trinary = FALSE; /* binary - not trinary or special */
        t->array = 0;       /* set not array (1 = indexed, 2 = whole array */
        t->targ = TINT;     /* set for integer arguments */
        t->trel = TINT;     /* set for integer result */
        t->eop = 0;         /* no corresponding floating or integer operator */
        t->seop = 0;        /* no corresponding string operator */
        t->xeop = 0;        /* no corresponding TXTYPE operator */
        t->iad = 0;         /* no corresponding integer operand */
        t->bad = 0;         /* no corresponding byte operand */
        t->sad = 0;         /* no corresponding string operand */
        t->xad = 0;         /* no corresponding TXTYPE operand */
        t->sgenv = 0;       /* no additional bytes */   
        t->endop = FALSE;   /* not end operator (ENDI, ENDF, etc) */
        t->local = -1;      /* neither global nor local */
        t->isaddr = 0;      /* not assign address */
        t->cgen68K = -1;    /* no 68000 code-generator info yet */
		t->cgenPPC = -1;    /* no PPC   code-generator info yet */
        t->txop = FALSE;    /* not legal on any TXTYPE */
    } /* for */

    xtP[TFLOAT].trel = TFLOAT;
    xtP[TFLOAT].op = FALSE;
    xtP[TINT].trel = TINT;
    xtP[TINT].op = FALSE;
    xtP[TMARK].trel = TMARK;
    xtP[TMARK].op = FALSE;
    xtP[TBYTE].trel = TBYTE;
    xtP[TBYTE].op = FALSE;
    xtP[TXTYPE].op = FALSE;
    xtP[TXTYPE].trel = TXTYPE;
    xtP[FARRAY].trel = TFLOAT; 
    xtP[FARRAY].op = FALSE;
    xtP[FARRAY].array = 2;
    xtP[FARRAY].sgenv = -2;
    xtP[IARRAY].trel = TINT; 
    xtP[IARRAY].op = FALSE;
    xtP[IARRAY].array = 2;
    xtP[IARRAY].sgenv = -2;
    xtP[BARRAY].trel = TBYTE; 
    xtP[BARRAY].op = FALSE;
    xtP[BARRAY].array = 2;
    xtP[BARRAY].sgenv = -2;
    xtP[MARRAY].trel = TMARK; 
    xtP[MARRAY].op = FALSE;
    xtP[MARRAY].array = 2;
    xtP[MARRAY].sgenv = -2;
    xtP[XARRAY].trel = TXTYPE; 
    xtP[XARRAY].op = FALSE;
    xtP[XARRAY].array = 2;
    xtP[XARRAY].sgenv = -2;
    xtP[ENDF].trel = TFLOAT; 
    xtP[ENDF].targ = TFLOAT; 
    xtP[ENDF].unary = TRUE;
    xtP[ENDF].endop = TRUE;
    xtP[ENDI].unary = TRUE; 
    xtP[ENDI].endop = TRUE;
    xtP[ENDI].txop = TRUE;
    xtP[ENDM].trel = TMARK; 
    xtP[ENDM].targ = TMARK; 
    xtP[ENDM].unary = TRUE;
    xtP[ENDM].endop = TRUE;
    xtP[FLITERAL].trel = TFLOAT;        
    xtP[FLITERAL].op = FALSE;
    xtP[FLITERAL].sgenv = -3;
    xtP[ILITERAL].sgenv = 4;
    xtP[ILITERAL].op = FALSE;
    xtP[MLITERAL].trel = TMARK;
    xtP[MLITERAL].op = FALSE;       
    xtP[MLITERAL].sgenv = 4;
    xtP[MLITC].trel = TMARK;
    xtP[MLITC].op = FALSE;      
    xtP[MLITC].sgenv = -6;
    xtP[SYSVAR].trel = TFLOAT;
    xtP[SYSVAR].op = FALSE;     
    xtP[SYSVAR].sgenv = 2;
    xtP[ISYSVAR].op = FALSE;
    xtP[ISYSVAR].sgenv = 2;
    xtP[MSYSVAR].op = FALSE;
    xtP[MSYSVAR].sgenv = 2; 
    xtP[MSYSVAR].trel = TMARK;
    xtP[MSYSVARA].op = FALSE;
    xtP[MSYSVARA].sgenv = 2;    
    xtP[MSYSVARA].trel = TMARK;
    xtP[MSYSVARA].isaddr = TRUE;
    xtP[TSYSVAR].op = FALSE;
    xtP[TSYSVAR].sgenv = 2;
    xtP[TSYSVAR].trel = TXTYPE;
    xtP[XGLOBALVAL].trel = TXTYPE;
    xtP[XGLOBALVAL].op = FALSE;
    xtP[XGLOBALVAL].sgenv = 4;
    xtP[XGLOBALVAL].local = 0;
    xtP[XLOCALVAL].trel = TXTYPE;
    xtP[XLOCALVAL].op = FALSE;
    xtP[XLOCALVAL].sgenv = 4;
    xtP[XLOCALVAL].local = 1;
    xtP[XPASSVAL].trel = TXTYPE;
    xtP[XPASSVAL].op = FALSE;
    xtP[XPASSVAL].sgenv = 4;
    xtP[XPASSVAL].local = 2;
    xtP[XGARRAYVAL].trel = TXTYPE;
    xtP[XGARRAYVAL].op = FALSE;
    xtP[XGARRAYVAL].sgenv = -1;
    xtP[XGARRAYVAL].array = 1;
    xtP[XGARRAYVAL].local = 0;
    xtP[XLARRAYVAL].trel = TXTYPE;
    xtP[XLARRAYVAL].op = FALSE;
    xtP[XLARRAYVAL].sgenv = -1;
    xtP[XLARRAYVAL].array = 1;
    xtP[XLARRAYVAL].local = 1;
    xtP[XGARRAY].trel = TXTYPE;
    xtP[XGARRAY].op = FALSE;
    xtP[XGARRAY].sgenv = -2;            
    xtP[XGARRAY].array = 2;
    xtP[XGARRAY].local = 0;
    xtP[XLARRAY].trel = TXTYPE;
    xtP[XLARRAY].op = FALSE;
    xtP[XLARRAY].sgenv = -2;            
    xtP[XLARRAY].array = 2;
    xtP[GLOBALVAL].op = FALSE;
    xtP[GLOBALVAL].trel = TFLOAT;   
    xtP[GLOBALVAL].sgenv = 4;
    xtP[GLOBALVAL].iad = IGLOBALVAL;
    xtP[GLOBALVAL].bad = BGLOBALVAL;
    xtP[GLOBALVAL].sad = MGLOBALVAL;
    xtP[GLOBALVAL].xad = XGLOBALVAL;
    xtP[GLOBALVAL].local = 0;
    xtP[IGLOBALVAL].op = FALSE;
    xtP[IGLOBALVAL].sgenv = 4;
    xtP[IGLOBALVAL].local = 0;
    xtP[BGLOBALVAL].op = FALSE;
    xtP[BGLOBALVAL].sgenv = 4;      
    xtP[BGLOBALVAL].trel = TBYTE;
    xtP[BGLOBALVAL].local = 0;
    xtP[MGLOBALVAL].op = FALSE;
    xtP[MGLOBALVAL].sgenv = 4;  
    xtP[MGLOBALVAL].trel = TMARK;
    xtP[MGLOBALVAL].local = 0;
    xtP[GLOBALADDR].op = FALSE;
    xtP[GLOBALADDR].trel = TFLOAT;  
    xtP[GLOBALADDR].sgenv = 4;
    xtP[GLOBALADDR].iad = IGLOBALADDR;
    xtP[GLOBALADDR].bad = BGLOBALADDR;
    xtP[GLOBALADDR].sad = MGLOBALADDR;
    xtP[GLOBALADDR].local = 0;
    xtP[GLOBALADDR].isaddr = TRUE;
    xtP[IGLOBALADDR].op = FALSE;
    xtP[IGLOBALADDR].sgenv = 4; 
    xtP[IGLOBALADDR].local = 0;
    xtP[IGLOBALADDR].isaddr = TRUE;
    xtP[BGLOBALADDR].op = FALSE; 
    xtP[BGLOBALADDR].sgenv = 4;     
    xtP[BGLOBALADDR].trel = TBYTE;
    xtP[BGLOBALADDR].local = 0;
    xtP[BGLOBALADDR].isaddr = TRUE;
    xtP[MGLOBALADDR].op = FALSE;
    xtP[MGLOBALADDR].trel = TMARK;  
    xtP[MGLOBALADDR].sgenv = 4;
    xtP[MGLOBALADDR].local = 0;
    xtP[MGLOBALADDR].isaddr = TRUE;
    xtP[LOCALVAL].op = FALSE;
    xtP[LOCALVAL].trel = TFLOAT;        
    xtP[LOCALVAL].sgenv = 4;
    xtP[LOCALVAL].iad = ILOCALVAL;  
    xtP[LOCALVAL].bad = BLOCALVAL;
    xtP[LOCALVAL].sad = MLOCALVAL;
    xtP[LOCALVAL].xad = XLOCALVAL;
    xtP[LOCALVAL].local = 1;
    xtP[ILOCALVAL].op = FALSE;
    xtP[ILOCALVAL].sgenv = 4;
    xtP[ILOCALVAL].local = 1;
    xtP[BLOCALVAL].op = FALSE;
    xtP[BLOCALVAL].sgenv = 4;       
    xtP[BLOCALVAL].trel = TBYTE;
    xtP[BLOCALVAL].local = 1;
    xtP[MLOCALVAL].op = FALSE;
    xtP[MLOCALVAL].trel = TMARK;    
    xtP[MLOCALVAL].sgenv = 4;
    xtP[MLOCALVAL].local = 1;
    xtP[LOCALADDR].op = FALSE;
    xtP[LOCALADDR].trel = TFLOAT;   
    xtP[LOCALADDR].sgenv = 4;
    xtP[LOCALADDR].iad = ILOCALADDR;
    xtP[LOCALADDR].bad = BLOCALADDR;
    xtP[LOCALADDR].sad = MLOCALADDR;
    xtP[LOCALADDR].isaddr = TRUE;
    xtP[LOCALADDR].local = 1;
    xtP[ILOCALADDR].op = FALSE;
    xtP[ILOCALADDR].sgenv = 4;
    xtP[ILOCALADDR].local = 1;
    xtP[ILOCALADDR].isaddr = TRUE;
    xtP[BLOCALADDR].op = FALSE;
    xtP[BLOCALADDR].sgenv = 4;      
    xtP[BLOCALADDR].trel = TBYTE;
    xtP[BLOCALADDR].local = 1;
    xtP[BLOCALADDR].isaddr = TRUE;
    xtP[MLOCALADDR].op = FALSE;
    xtP[MLOCALADDR].trel = TMARK;   
    xtP[MLOCALADDR].sgenv = 4;
    xtP[MLOCALADDR].local = 1;
    xtP[MLOCALADDR].isaddr = TRUE;
    xtP[PASSVAL].op = FALSE;
    xtP[PASSVAL].trel = TFLOAT;     
    xtP[PASSVAL].sgenv = 4;
    xtP[PASSVAL].iad = IPASSVAL;
    xtP[PASSVAL].bad = BPASSVAL;
    xtP[PASSVAL].sad = MPASSVAL;
    xtP[PASSVAL].xad = XPASSVAL;
    xtP[PASSVAL].local = 2;
    xtP[IPASSVAL].op = FALSE;
    xtP[IPASSVAL].sgenv = 4;
    xtP[IPASSVAL].local = 2;
    xtP[BPASSVAL].op = FALSE;
    xtP[BPASSVAL].sgenv = 4;        
    xtP[BPASSVAL].trel = TBYTE;
    xtP[BPASSVAL].local = 2;
    xtP[MPASSVAL].op = FALSE;
    xtP[MPASSVAL].trel = TMARK; 
    xtP[MPASSVAL].sgenv = 4;
    xtP[MPASSVAL].local = 2;
    xtP[PASSADDR].op = FALSE;
    xtP[PASSADDR].trel = TFLOAT;    
    xtP[PASSADDR].sgenv = 4;
    xtP[PASSADDR].iad = IPASSADDR;
    xtP[PASSADDR].bad = BPASSADDR;
    xtP[PASSADDR].sad = MPASSADDR;
    xtP[PASSADDR].local = 2;
    xtP[PASSADDR].isaddr = TRUE;
    xtP[IPASSADDR].op = FALSE;
    xtP[IPASSADDR].sgenv = 4;
    xtP[IPASSADDR].local = 2;
    xtP[IPASSADDR].isaddr = TRUE;
    xtP[BPASSADDR].op = FALSE;
    xtP[BPASSADDR].sgenv = 4;       
    xtP[BPASSADDR].trel = TBYTE;
    xtP[BPASSADDR].local = 2;
    xtP[BPASSADDR].isaddr = TRUE;
    xtP[MPASSADDR].op = FALSE;
    xtP[MPASSADDR].sgenv = 4;       
    xtP[MPASSADDR].trel = TMARK;
    xtP[MPASSADDR].local = 2;
    xtP[MPASSADDR].isaddr = TRUE;
    xtP[GARRAYVAL].op = FALSE;
    xtP[GARRAYVAL].trel = TFLOAT;   
    xtP[GARRAYVAL].sgenv = -1;
    xtP[GARRAYVAL].array = 1;
    xtP[GARRAYVAL].iad = IGARRAYVAL;
    xtP[GARRAYVAL].bad = BGARRAYVAL;
    xtP[GARRAYVAL].sad = MGARRAYVAL;
    xtP[GARRAYVAL].xad = XGARRAYVAL;
    xtP[GARRAYVAL].local = 0;
    xtP[IGARRAYVAL].op = FALSE;
    xtP[IGARRAYVAL].sgenv = -1;
    xtP[IGARRAYVAL].array = 1;
    xtP[IGARRAYVAL].local = 0;
    xtP[BGARRAYVAL].op = FALSE;
    xtP[BGARRAYVAL].sgenv = -1;     
    xtP[BGARRAYVAL].trel = TBYTE;
    xtP[BGARRAYVAL].array = 1;
    xtP[BGARRAYVAL].local = 0;
    xtP[MGARRAYVAL].op = FALSE;
    xtP[MGARRAYVAL].trel = TMARK;   
    xtP[MGARRAYVAL].sgenv = -1;
    xtP[MGARRAYVAL].array = 1;
    xtP[MGARRAYVAL].local = 0;
    xtP[GARRAYADDR].op = FALSE;
    xtP[GARRAYADDR].trel = TFLOAT;  
    xtP[GARRAYADDR].sgenv = -1;
    xtP[GARRAYADDR].array = 1;
    xtP[GARRAYADDR].iad = IGARRAYADDR;
    xtP[GARRAYADDR].bad = BGARRAYADDR;
    xtP[GARRAYADDR].sad = MGARRAYADDR;
    xtP[GARRAYADDR].local = 0;
    xtP[GARRAYADDR].isaddr = TRUE;
    xtP[IGARRAYADDR].sgenv = -1;
    xtP[IGARRAYADDR].op = FALSE;
    xtP[IGARRAYADDR].array = 1;
    xtP[IGARRAYADDR].local = 0;
    xtP[IGARRAYADDR].isaddr = TRUE;
    xtP[BGARRAYADDR].op = FALSE;
    xtP[BGARRAYADDR].sgenv = -1;        
    xtP[BGARRAYADDR].trel = TBYTE;
    xtP[BGARRAYADDR].array = 1;
    xtP[BGARRAYADDR].local = 0;
    xtP[BGARRAYADDR].isaddr = TRUE;
    xtP[MGARRAYADDR].op = FALSE;
    xtP[MGARRAYADDR].trel = TMARK;  
    xtP[MGARRAYADDR].sgenv = -1;
    xtP[MGARRAYADDR].array = 1;
    xtP[MGARRAYADDR].local = 0;
    xtP[MGARRAYADDR].isaddr = TRUE;
    xtP[LARRAYVAL].op = FALSE;
    xtP[LARRAYVAL].trel = TFLOAT;   
    xtP[LARRAYVAL].sgenv = -1;
    xtP[LARRAYVAL].array = 1;
    xtP[LARRAYVAL].iad = ILARRAYVAL;
    xtP[LARRAYVAL].bad = BLARRAYVAL;
    xtP[LARRAYVAL].sad = MLARRAYVAL;
    xtP[LARRAYVAL].xad = XLARRAYVAL;
    xtP[LARRAYVAL].local = 1;
    xtP[ILARRAYVAL].op = FALSE;
    xtP[ILARRAYVAL].sgenv = -1;
    xtP[ILARRAYVAL].array = 1;
    xtP[ILARRAYVAL].local = 1;
    xtP[BLARRAYVAL].op = FALSE;
    xtP[BLARRAYVAL].sgenv = -1;     
    xtP[BLARRAYVAL].trel = TBYTE;
    xtP[BLARRAYVAL].array = 1;
    xtP[BLARRAYVAL].local = 1;
    xtP[MLARRAYVAL].op = FALSE;
    xtP[MLARRAYVAL].trel = TMARK;   
    xtP[MLARRAYVAL].sgenv = -1;
    xtP[MLARRAYVAL].array = 1;
    xtP[MLARRAYVAL].local = 1;
    xtP[LARRAYADDR].op = FALSE;
    xtP[LARRAYADDR].trel = TFLOAT;  
    xtP[LARRAYADDR].sgenv = -1;
    xtP[LARRAYADDR].array = 1;
    xtP[LARRAYADDR].iad = ILARRAYADDR;
    xtP[LARRAYADDR].bad = BLARRAYADDR;
    xtP[LARRAYADDR].sad = MLARRAYADDR;
    xtP[LARRAYADDR].local = 1;
    xtP[LARRAYADDR].isaddr = TRUE;
    xtP[ILARRAYADDR].op = FALSE;
    xtP[ILARRAYADDR].sgenv = -1;
    xtP[ILARRAYADDR].array = 1;
    xtP[ILARRAYADDR].local = 1;
    xtP[ILARRAYADDR].isaddr = TRUE;
    xtP[BLARRAYADDR].op = FALSE;
    xtP[BLARRAYADDR].sgenv = -1;        
    xtP[BLARRAYADDR].trel = TBYTE;
    xtP[BLARRAYADDR].array = 1;
    xtP[BLARRAYADDR].local = 1;
    xtP[BLARRAYADDR].isaddr = TRUE;
    xtP[MLARRAYADDR].op = FALSE;
    xtP[MLARRAYADDR].trel = TMARK;  
    xtP[MLARRAYADDR].sgenv = -1;
    xtP[MLARRAYADDR].array = 1;
    xtP[MLARRAYADDR].local = 1;
    xtP[MLARRAYADDR].isaddr = TRUE;
    xtP[PARRAYVAL].op = FALSE;
    xtP[PARRAYVAL].trel = TFLOAT;   
    xtP[PARRAYVAL].sgenv = -1;
    xtP[PARRAYVAL].array = 1;
    xtP[PARRAYVAL].iad = IPARRAYVAL;
    xtP[PARRAYVAL].bad = BPARRAYVAL;
    xtP[PARRAYVAL].sad = MPARRAYVAL;
    xtP[PARRAYVAL].xad = XPARRAYVAL;
    xtP[PARRAYVAL].local = 2;
    xtP[IPARRAYVAL].op = FALSE;
    xtP[IPARRAYVAL].sgenv = -1;
    xtP[IPARRAYVAL].array = 1;
    xtP[IPARRAYVAL].local = 2;
    xtP[BPARRAYVAL].op = FALSE;
    xtP[BPARRAYVAL].sgenv = -1;     
    xtP[BPARRAYVAL].trel = TBYTE;
    xtP[BPARRAYVAL].array = 1;
    xtP[BPARRAYVAL].local = 2;
    xtP[MPARRAYVAL].op = FALSE;
    xtP[MPARRAYVAL].trel = TMARK;   
    xtP[MPARRAYVAL].sgenv = -1;
    xtP[MPARRAYVAL].array = 1;
    xtP[MPARRAYVAL].local = 2;
    xtP[XPARRAYVAL].trel = TXTYPE;
    xtP[XPARRAYVAL].op = FALSE;
    xtP[XPARRAYVAL].sgenv = -1;
    xtP[XPARRAYVAL].array = 1;
    xtP[XPARRAYVAL].local = 2;
    xtP[PARRAYADDR].op = FALSE;
    xtP[PARRAYADDR].trel = TFLOAT;  
    xtP[PARRAYADDR].sgenv = -1;
    xtP[PARRAYADDR].array = 1;
    xtP[PARRAYADDR].iad = IPARRAYADDR;
    xtP[PARRAYADDR].bad = BPARRAYADDR;
    xtP[PARRAYADDR].sad = MPARRAYADDR;
    xtP[PARRAYADDR].local = 2;
    xtP[PARRAYADDR].isaddr = TRUE;
    xtP[IPARRAYADDR].op = FALSE;
    xtP[IPARRAYADDR].sgenv = -1;
    xtP[IPARRAYADDR].array = 1;
    xtP[IPARRAYADDR].local = 2;
    xtP[IPARRAYADDR].isaddr = TRUE;
    xtP[BPARRAYADDR].op = FALSE;
    xtP[BPARRAYADDR].sgenv = -1;        
    xtP[BPARRAYADDR].trel = TBYTE;
    xtP[BPARRAYADDR].array = 1;
    xtP[BPARRAYADDR].local = 2;
    xtP[BPARRAYADDR].isaddr = TRUE;
    xtP[MPARRAYADDR].op = FALSE;
    xtP[MPARRAYADDR].trel = TMARK;  
    xtP[MPARRAYADDR].sgenv = -1;
    xtP[MPARRAYADDR].array = 1;
    xtP[MPARRAYADDR].local = 2;
    xtP[MPARRAYADDR].isaddr = TRUE;
    xtP[GDYARRAYVAL].op = FALSE;
    xtP[GDYARRAYVAL].trel = TFLOAT; 
    xtP[GDYARRAYVAL].sgenv = -1;
    xtP[GDYARRAYVAL].array = 1;
    xtP[GDYARRAYVAL].iad = IGDYARRAYVAL;
    xtP[GDYARRAYVAL].bad = BGDYARRAYVAL;
    xtP[GDYARRAYVAL].sad = MGDYARRAYVAL;
    xtP[GDYARRAYVAL].xad = XGDYARRAYVAL;
    xtP[GDYARRAYVAL].local = 3;
    xtP[IGDYARRAYVAL].op = FALSE;
    xtP[IGDYARRAYVAL].sgenv = -1;
    xtP[IGDYARRAYVAL].array = 1;
    xtP[IGDYARRAYVAL].local = 3;
    xtP[BGDYARRAYVAL].op = FALSE;
    xtP[BGDYARRAYVAL].sgenv = -1;       
    xtP[BGDYARRAYVAL].trel = TBYTE;
    xtP[BGDYARRAYVAL].array = 1;
    xtP[BGDYARRAYVAL].local = 3;
    xtP[MGDYARRAYVAL].op = FALSE;
    xtP[MGDYARRAYVAL].trel = TMARK; 
    xtP[MGDYARRAYVAL].sgenv = -1;
    xtP[MGDYARRAYVAL].array = 1;
    xtP[MGDYARRAYVAL].local = 3;
    xtP[XGDYARRAYVAL].trel = TXTYPE;
    xtP[XGDYARRAYVAL].op = FALSE;
    xtP[XGDYARRAYVAL].sgenv = -1;
    xtP[XGDYARRAYVAL].array = 1;
    xtP[XGDYARRAYVAL].local = 3;
    xtP[GDYARRAYADDR].op = FALSE;
    xtP[GDYARRAYADDR].trel = TFLOAT;    
    xtP[GDYARRAYADDR].sgenv = -1;
    xtP[GDYARRAYADDR].array = 1;
    xtP[GDYARRAYADDR].iad = IGDYARRAYADDR;
    xtP[GDYARRAYADDR].bad = BGDYARRAYADDR;
    xtP[GDYARRAYADDR].sad = MGDYARRAYADDR;
    xtP[GDYARRAYADDR].local = 3;
    xtP[GDYARRAYADDR].isaddr = TRUE;
    xtP[IGDYARRAYADDR].op = FALSE;
    xtP[IGDYARRAYADDR].sgenv = -1;
    xtP[IGDYARRAYADDR].array = 1;
    xtP[IGDYARRAYADDR].local = 3;
    xtP[IGDYARRAYADDR].isaddr = TRUE;
    xtP[BGDYARRAYADDR].op = FALSE;
    xtP[BGDYARRAYADDR].sgenv = -1;      
    xtP[BGDYARRAYADDR].trel = TBYTE;
    xtP[BGDYARRAYADDR].array = 1;
    xtP[BGDYARRAYADDR].local = 3;
    xtP[BGDYARRAYADDR].isaddr = TRUE;
    xtP[MGDYARRAYADDR].op = FALSE;
    xtP[MGDYARRAYADDR].trel = TMARK;    
    xtP[MGDYARRAYADDR].sgenv = -1;
    xtP[MGDYARRAYADDR].array = 1;
    xtP[MGDYARRAYADDR].local = 3;
    xtP[MGDYARRAYADDR].isaddr = TRUE;
    xtP[LDYARRAYVAL].op = FALSE;
    xtP[LDYARRAYVAL].trel = TFLOAT; 
    xtP[LDYARRAYVAL].sgenv = -1;
    xtP[LDYARRAYVAL].array = 1;
    xtP[LDYARRAYVAL].iad = ILDYARRAYVAL;
    xtP[LDYARRAYVAL].bad = BLDYARRAYVAL;
    xtP[LDYARRAYVAL].sad = MLDYARRAYVAL;
    xtP[LDYARRAYVAL].xad = XLDYARRAYVAL;
    xtP[LDYARRAYVAL].local = 4;
    xtP[ILDYARRAYVAL].op = FALSE;
    xtP[ILDYARRAYVAL].sgenv = -1;
    xtP[ILDYARRAYVAL].array = 1;
    xtP[ILDYARRAYVAL].local = 4;
    xtP[BLDYARRAYVAL].op = FALSE;
    xtP[BLDYARRAYVAL].sgenv = -1;       
    xtP[BLDYARRAYVAL].trel = TBYTE;
    xtP[BLDYARRAYVAL].array = 1;
    xtP[BLDYARRAYVAL].local = 4;
    xtP[MLDYARRAYVAL].op = FALSE;
    xtP[MLDYARRAYVAL].trel = TMARK; 
    xtP[MLDYARRAYVAL].sgenv = -1;
    xtP[MLDYARRAYVAL].array = 1;
    xtP[MLDYARRAYVAL].local = 4;
    xtP[XLDYARRAYVAL].trel = TXTYPE;
    xtP[XLDYARRAYVAL].op = FALSE;
    xtP[XLDYARRAYVAL].sgenv = -1;
    xtP[XLDYARRAYVAL].array = 1;
    xtP[XLDYARRAYVAL].local = 4;
    xtP[LDYARRAYADDR].op = FALSE;
    xtP[LDYARRAYADDR].trel = TFLOAT;    
    xtP[LDYARRAYADDR].sgenv = -1;
    xtP[LDYARRAYADDR].array = 1;
    xtP[LDYARRAYADDR].iad = ILDYARRAYADDR;
    xtP[LDYARRAYADDR].bad = BLDYARRAYADDR;
    xtP[LDYARRAYADDR].sad = MLDYARRAYADDR;
    xtP[LDYARRAYADDR].local = 4;
    xtP[LDYARRAYADDR].isaddr = TRUE;
    xtP[ILDYARRAYADDR].op = FALSE;
    xtP[ILDYARRAYADDR].sgenv = -1;
    xtP[ILDYARRAYADDR].array = 1;
    xtP[ILDYARRAYADDR].local = 4;
    xtP[ILDYARRAYADDR].isaddr = TRUE;
    xtP[BLDYARRAYADDR].op = FALSE;
    xtP[BLDYARRAYADDR].sgenv = -1;      
    xtP[BLDYARRAYADDR].trel = TBYTE;
    xtP[BLDYARRAYADDR].array = 1;
    xtP[BLDYARRAYADDR].local = 4;
    xtP[BLDYARRAYADDR].isaddr = TRUE;
    xtP[MLDYARRAYADDR].op = FALSE;
    xtP[MLDYARRAYADDR].trel = TMARK;    
    xtP[MLDYARRAYADDR].sgenv = -1;
    xtP[MLDYARRAYADDR].array = 1;
    xtP[MLDYARRAYADDR].local = 4;
    xtP[MLDYARRAYADDR].isaddr = TRUE;
    xtP[PDYARRAYVAL].op = FALSE;
    xtP[PDYARRAYVAL].trel = TFLOAT; 
    xtP[PDYARRAYVAL].sgenv = -1;
    xtP[PDYARRAYVAL].array = 1;
    xtP[PDYARRAYVAL].iad = IPDYARRAYVAL;
    xtP[PDYARRAYVAL].bad = BPDYARRAYVAL;
    xtP[PDYARRAYVAL].sad = MPDYARRAYVAL;
    xtP[PDYARRAYVAL].xad = XPDYARRAYVAL;
    xtP[PDYARRAYVAL].local = 5;
    xtP[IPDYARRAYVAL].op = FALSE;
    xtP[IPDYARRAYVAL].sgenv = -1;
    xtP[IPDYARRAYVAL].array = 1;
    xtP[IPDYARRAYVAL].local = 5;
    xtP[BPDYARRAYVAL].op = FALSE;
    xtP[BPDYARRAYVAL].sgenv = -1;       
    xtP[BPDYARRAYVAL].trel = TBYTE;
    xtP[BPDYARRAYVAL].array = 1;
    xtP[BPDYARRAYVAL].local = 5;
    xtP[MPDYARRAYVAL].op = FALSE;
    xtP[MPDYARRAYVAL].trel = TMARK; 
    xtP[MPDYARRAYVAL].sgenv = -1;
    xtP[MPDYARRAYVAL].array = 1;
    xtP[MPDYARRAYVAL].local = 5;
    xtP[XPDYARRAYVAL].trel = TXTYPE;
    xtP[XPDYARRAYVAL].op = FALSE;
    xtP[XPDYARRAYVAL].sgenv = -1;
    xtP[XPDYARRAYVAL].array = 1;
    xtP[XPDYARRAYVAL].local = 5;
    xtP[PDYARRAYADDR].op = FALSE;
    xtP[PDYARRAYADDR].trel = TFLOAT;    
    xtP[PDYARRAYADDR].sgenv = -1;
    xtP[PDYARRAYADDR].array = 1;
    xtP[PDYARRAYADDR].iad = IPDYARRAYADDR;
    xtP[PDYARRAYADDR].bad = BPDYARRAYADDR;
    xtP[PDYARRAYADDR].sad = MPDYARRAYADDR;
    xtP[PDYARRAYADDR].local = 5;
    xtP[PDYARRAYADDR].isaddr = TRUE;
    xtP[IPDYARRAYADDR].op = FALSE;
    xtP[IPDYARRAYADDR].sgenv = -1;
    xtP[IPDYARRAYADDR].array = 1;
    xtP[IPDYARRAYADDR].local = 5;
    xtP[IPDYARRAYADDR].isaddr = TRUE;
    xtP[BPDYARRAYADDR].op = FALSE;
    xtP[BPDYARRAYADDR].sgenv = -1;      
    xtP[BPDYARRAYADDR].trel = TBYTE;
    xtP[BPDYARRAYADDR].array = 1;
    xtP[BPDYARRAYADDR].local = 5;
    xtP[BPDYARRAYADDR].isaddr = TRUE;
    xtP[MPDYARRAYADDR].op = FALSE;
    xtP[MPDYARRAYADDR].trel = TMARK;    
    xtP[MPDYARRAYADDR].sgenv = -1;
    xtP[MPDYARRAYADDR].array = 1;
    xtP[MPDYARRAYADDR].local = 5;
    xtP[MPDYARRAYADDR].isaddr = TRUE;
    xtP[GARRAY].op = FALSE;
    xtP[GARRAY].trel = TFLOAT;      
    xtP[GARRAY].sgenv = -2;
    xtP[GARRAY].iad = IGARRAY;      
    xtP[GARRAY].array = 2;  
    xtP[GARRAY].bad = BGARRAY;
    xtP[GARRAY].sad = MGARRAY;
    xtP[GARRAY].xad = XGARRAY;
    xtP[GARRAY].local = 0;
    xtP[IGARRAY].op = FALSE;
    xtP[IGARRAY].sgenv = -2;            
    xtP[IGARRAY].array = 2;
    xtP[IGARRAY].local = 0;
    xtP[BGARRAY].op = FALSE;
    xtP[BGARRAY].sgenv = -2;        
    xtP[BGARRAY].trel = TBYTE;
    xtP[BGARRAY].array = 2;
    xtP[BGARRAY].local = 0;
    xtP[MGARRAY].op = FALSE;
    xtP[MGARRAY].trel = TMARK;      
    xtP[MGARRAY].sgenv = -2;
    xtP[MGARRAY].array = 2;
    xtP[MGARRAY].local = 0;
    xtP[LARRAY].op = FALSE;
    xtP[LARRAY].trel = TFLOAT;      
    xtP[LARRAY].sgenv = -2;
    xtP[LARRAY].array = 2;
    xtP[LARRAY].iad = ILARRAY;
    xtP[LARRAY].bad = BLARRAY;
    xtP[LARRAY].sad = MLARRAY;
    xtP[LARRAY].xad = XLARRAY;
    xtP[LARRAY].local = 1;
    xtP[ILARRAY].op = FALSE;
    xtP[ILARRAY].sgenv = -2;            
    xtP[ILARRAY].array = 2;
    xtP[ILARRAY].local = 1;
    xtP[BLARRAY].op = FALSE;
    xtP[BLARRAY].sgenv = -2;            
    xtP[BLARRAY].trel = TBYTE;
    xtP[BLARRAY].array = 2;
    xtP[BLARRAY].local = 1;
    xtP[MLARRAY].op = FALSE;
    xtP[MLARRAY].trel = TMARK;      
    xtP[MLARRAY].sgenv = -2;
    xtP[MLARRAY].array = 2;
    xtP[MLARRAY].local = 1;
    xtP[PARRAY].op = FALSE;
    xtP[PARRAY].trel = TFLOAT;      
    xtP[PARRAY].sgenv = -2;
    xtP[PARRAY].iad = IPARRAY;      
    xtP[PARRAY].array = 2;
    xtP[PARRAY].bad = BPARRAY;
    xtP[PARRAY].sad = MPARRAY;
    xtP[PARRAY].xad = XPARRAY;
    xtP[PARRAY].local = 2;
    xtP[IPARRAY].op = FALSE;
    xtP[IPARRAY].sgenv = -2;            
    xtP[IPARRAY].array = 2;
    xtP[IPARRAY].local = 2;
    xtP[BPARRAY].op = FALSE;        
    xtP[BPARRAY].trel = TBYTE;
    xtP[BPARRAY].array = 2;
    xtP[BPARRAY].sgenv = -2;
    xtP[BPARRAY].local = 2;
    xtP[MPARRAY].op = FALSE;
    xtP[MPARRAY].trel = TMARK;      
    xtP[MPARRAY].sgenv = -2;
    xtP[MPARRAY].array = 2;
    xtP[MPARRAY].local = 2;
    xtP[XPARRAY].op = FALSE;
    xtP[XPARRAY].op = FALSE;
    xtP[XPARRAY].sgenv = -2;            
    xtP[XPARRAY].array = 2;
    xtP[XPARRAY].local = 2;
    xtP[GDYARRAY].op = FALSE;
    xtP[GDYARRAY].trel = TFLOAT;        
    xtP[GDYARRAY].sgenv = -2;
    xtP[GDYARRAY].iad = IGDYARRAY;      
    xtP[GDYARRAY].array = 2;
    xtP[GDYARRAY].bad = BGDYARRAY;
    xtP[GDYARRAY].sad = MGDYARRAY;
    xtP[GDYARRAY].xad = XGDYARRAY;
    xtP[GDYARRAY].local = 3;
    xtP[IGDYARRAY].op = FALSE;
    xtP[IGDYARRAY].sgenv = -2;          
    xtP[IGDYARRAY].array = 2;
    xtP[IGDYARRAY].local = 3;
    xtP[BGDYARRAY].op = FALSE;      
    xtP[BGDYARRAY].trel = TBYTE;
    xtP[BGDYARRAY].array = 2;
    xtP[BGDYARRAY].sgenv = -2;
    xtP[BGDYARRAY].local = 3;
    xtP[MGDYARRAY].op = FALSE;
    xtP[MGDYARRAY].trel = TMARK;        
    xtP[MGDYARRAY].sgenv = -2;
    xtP[MGDYARRAY].array = 2;
    xtP[MGDYARRAY].local = 3;
    xtP[XGDYARRAY].op = FALSE;
    xtP[XGDYARRAY].op = FALSE;
    xtP[XGDYARRAY].sgenv = -2;          
    xtP[XGDYARRAY].array = 2;
    xtP[XGDYARRAY].local = 3;
    xtP[LDYARRAY].op = FALSE;
    xtP[LDYARRAY].trel = TFLOAT;        
    xtP[LDYARRAY].sgenv = -2;
    xtP[LDYARRAY].iad = ILDYARRAY;      
    xtP[LDYARRAY].array = 2;
    xtP[LDYARRAY].bad = BLDYARRAY;
    xtP[LDYARRAY].sad = MLDYARRAY;
    xtP[LDYARRAY].xad = XLDYARRAY;
    xtP[LDYARRAY].local = 4;
    xtP[ILDYARRAY].op = FALSE;
    xtP[ILDYARRAY].sgenv = -2;          
    xtP[ILDYARRAY].array = 2;
    xtP[ILDYARRAY].local = 4;
    xtP[BLDYARRAY].op = FALSE;      
    xtP[BLDYARRAY].trel = TBYTE;
    xtP[BLDYARRAY].array = 2;
    xtP[BLDYARRAY].sgenv = -2;
    xtP[BLDYARRAY].local = 4;
    xtP[MLDYARRAY].op = FALSE;
    xtP[MLDYARRAY].trel = TMARK;        
    xtP[MLDYARRAY].sgenv = -2;
    xtP[MLDYARRAY].array = 2;
    xtP[MLDYARRAY].local = 4;
    xtP[XLDYARRAY].op = FALSE;
    xtP[XLDYARRAY].op = FALSE;
    xtP[XLDYARRAY].sgenv = -2;          
    xtP[XLDYARRAY].array = 2;
    xtP[XLDYARRAY].local = 4;
    xtP[PDYARRAY].op = FALSE;
    xtP[PDYARRAY].trel = TFLOAT;        
    xtP[PDYARRAY].sgenv = -2;
    xtP[PDYARRAY].iad = IPDYARRAY;      
    xtP[PDYARRAY].array = 2;
    xtP[PDYARRAY].bad = BPDYARRAY;
    xtP[PDYARRAY].sad = MPDYARRAY;
    xtP[PDYARRAY].xad = XPDYARRAY;
    xtP[PDYARRAY].local = 5;
    xtP[IPDYARRAY].op = FALSE;
    xtP[IPDYARRAY].sgenv = -2;          
    xtP[IPDYARRAY].array = 2;
    xtP[IPDYARRAY].local = 5;
    xtP[BPDYARRAY].op = FALSE;      
    xtP[BPDYARRAY].trel = TBYTE;
    xtP[BPDYARRAY].array = 2;
    xtP[BPDYARRAY].sgenv = -2;
    xtP[BPDYARRAY].local = 5;
    xtP[MPDYARRAY].op = FALSE;
    xtP[MPDYARRAY].trel = TMARK;        
    xtP[MPDYARRAY].sgenv = -2;
    xtP[MPDYARRAY].array = 2;
    xtP[MPDYARRAY].local = 5;
    xtP[XPDYARRAY].op = FALSE;
    xtP[XPDYARRAY].op = FALSE;
    xtP[XPDYARRAY].sgenv = -2;          
    xtP[XPDYARRAY].array = 2;
    xtP[XPDYARRAY].local = 5;
    xtP[LT].targ = TFLOAT;          
    xtP[LT].eop = ILT;      
    xtP[LT].seop = MLT;
    xtP[SPLT].targ = TFLOAT;        
    xtP[SPLT].eop = ISPLT;  
    xtP[SPLT].seop = MSPLT;
    xtP[SPLT].trel = TFLOAT;
    xtP[GT].targ = TFLOAT;          
    xtP[GT].eop = IGT;      
    xtP[GT].seop = MGT;
    xtP[SPGT].targ = TFLOAT;        
    xtP[SPGT].eop = ISPGT;  
    xtP[SPGT].seop = MSPGT;
    xtP[SPGT].trel = TFLOAT;
    xtP[LE].targ = TFLOAT;          
    xtP[LE].eop = ILE;      
    xtP[LE].seop = MLE;
    xtP[SPLE].targ = TFLOAT;        
    xtP[SPLE].eop = ISPLE;  
    xtP[SPLE].seop = MSPLE;
    xtP[SPLE].trel = TFLOAT;
    xtP[GE].targ = TFLOAT;          
    xtP[GE].eop = IGE;      
    xtP[GE].seop = MGE;
    xtP[SPGE].targ = TFLOAT;        
    xtP[SPGE].eop = ISPGE;  
    xtP[SPGE].seop = MSPGE;
    xtP[SPGE].trel = TFLOAT;
    xtP[EQ].targ = TFLOAT;          
    xtP[EQ].eop = IEQ;      
    xtP[EQ].seop = MEQ;
    xtP[EQ].xeop = TXEQ;
    xtP[SPEQ].targ = TFLOAT;        
    xtP[SPEQ].eop = ISPEQ;  
    xtP[SPEQ].seop = MSPEQ;
    xtP[SPEQ].trel = TFLOAT;
    xtP[NE].targ = TFLOAT;          
    xtP[NE].eop = INE;      
    xtP[NE].seop = MNE;
    xtP[NE].xeop = TXNEQ;
    xtP[SPNE].targ = TFLOAT;        
    xtP[SPNE].eop = ISPNE;  
    xtP[SPNE].seop = MSPNE;
    xtP[SPNE].trel = TFLOAT;
    xtP[ILT].eop = LT;
    xtP[ISPLT].eop = SPLT;
    xtP[IGT].eop = GT;
    xtP[ISPGT].eop = SPGT;
    xtP[ILE].eop = LE;
    xtP[ISPLE].eop = SPLE;
    xtP[IGE].eop = GE;
    xtP[ISPGE].eop = SPGE;
    xtP[IEQ].eop = EQ;
    xtP[ISPEQ].eop = SPEQ;
    xtP[INE].eop = NE;
    xtP[ISPNE].eop = SPNE;
    xtP[MLT].targ = TMARK;
    xtP[MSPLT].targ = TMARK;    
    xtP[MSPLT].trel = TMARK;
    xtP[MGT].targ = TMARK;
    xtP[MSPGT].targ = TMARK;    
    xtP[MSPGT].trel = TMARK;
    xtP[MLE].targ = TMARK;
    xtP[MSPLE].targ = TMARK;    
    xtP[MSPLE].trel = TMARK;
    xtP[MGE].targ = TMARK;
    xtP[MSPGE].targ = TMARK;    
    xtP[MSPGE].trel = TMARK;
    xtP[MEQ].targ = TMARK;
    xtP[MSPEQ].targ = TMARK;    
    xtP[MSPEQ].trel = TMARK;
    xtP[MNE].targ = TMARK;
    xtP[MSPNE].targ = TMARK;    
    xtP[MSPNE].trel = TMARK;
    xtP[TXEQ].targ = TINT;  
    xtP[TXEQ].trel = TINT;
    xtP[TXEQ].txop = TRUE;
    xtP[TXNEQ].targ = TINT; 
    xtP[TXNEQ].trel = TINT;
    xtP[TXNEQ].txop = TRUE;
    xtP[IBRANCHT].unary = TRUE;
    xtP[IBRANCHT].sgenv = 2;
    xtP[IBRANCHF].unary = TRUE;
    xtP[IBRANCHF].sgenv = 2;
    xtP[MCAT].targ = TMARK; 
    xtP[MCAT].trel = TMARK;
    xtP[ZK].op = FALSE;
    xtP[ZK].sgenv = 2;
    xtP[MOD].targ = TFLOAT;     
    xtP[MOD].trel = TFLOAT;
    xtP[ITOF].trel = TFLOAT;        
    xtP[ITOF].unary = TRUE;
    xtP[MUNIT].unary = TRUE;    
    xtP[MUNIT].targ = TMARK;
    xtP[FTOI].targ = TFLOAT;        
    xtP[FTOI].unary = TRUE;
    xtP[ASSIGN].targ = TFLOAT;  
    xtP[ASSIGN].trel = TFLOAT;
    xtP[MASSIGN].targ = TMARK;  
    xtP[MASSIGN].trel = TMARK;
    xtP[PLUS].targ = TFLOAT;    
    xtP[PLUS].trel = TFLOAT;        
    xtP[PLUS].eop = IPLUS;
    xtP[PLUS].seop = MCAT;
    xtP[IPLUS].eop = PLUS;
    xtP[IINC].unary = TRUE;
    xtP[MINUS].targ = TFLOAT;   
    xtP[MINUS].trel = TFLOAT;
    xtP[MINUS].eop = IMINUS;
    xtP[IMINUS].eop = MINUS;
    xtP[IDEC].unary = TRUE;
    xtP[TIMES].targ = TFLOAT;   
    xtP[TIMES].trel = TFLOAT;   
    xtP[TIMES].eop = ITIMES;
    xtP[ITIMES].eop = TIMES;
    xtP[DIVIDE].targ = TFLOAT;  
    xtP[DIVIDE].trel = TFLOAT;
    xtP[UMINUS].targ = TFLOAT;  
    xtP[UMINUS].trel = TFLOAT;
    xtP[UMINUS].unary = TRUE;   
    xtP[UMINUS].eop = IUMINUS;
    xtP[IUMINUS].eop = UMINUS;  
    xtP[IUMINUS].unary = TRUE;
    xtP[NOT].unary = TRUE;
    xtP[EXPONENT].targ = TFLOAT;    
    xtP[EXPONENT].trel = TFLOAT;
    xtP[EXPONIC].targ = TFLOAT;     
    xtP[EXPONIC].trel = TFLOAT;
    xtP[EXPONIC].sgenv = 2;         
    xtP[EXPONIC].unary = TRUE;
    xtP[ARCTAN].targ = TFLOAT;      
    xtP[ARCTAN].trel = TFLOAT;
    xtP[ARCTAN].unary = TRUE;
    xtP[ARCTAN2].targ = TFLOAT; 
    xtP[ARCTAN2].trel = TFLOAT;
    xtP[INT].unary = TRUE;      
    xtP[INT].targ = TFLOAT;
    xtP[INT].trel = TFLOAT;
    xtP[FRAC].unary = TRUE;     
    xtP[FRAC].targ = TFLOAT;    
    xtP[FRAC].trel = TFLOAT;
    xtP[ROUND].unary = TRUE;    
    xtP[ROUND].targ = TFLOAT;
    xtP[SIGN].targ = TFLOAT;
    xtP[SIGN].trel = TFLOAT;    
    xtP[SIGN].unary = TRUE;
    xtP[SIGN].eop = ISIGN;
    xtP[ISIGN].eop = SIGN;
    xtP[ISIGN].unary = TRUE;
    xtP[SIN].unary = TRUE;  
    xtP[SIN].targ = TFLOAT; 
    xtP[SIN].trel = TFLOAT;
    xtP[COS].unary = TRUE;  
    xtP[COS].targ = TFLOAT; 
    xtP[COS].trel = TFLOAT;
    xtP[TAN].unary = TRUE;  
    xtP[TAN].targ = TFLOAT; 
    xtP[TAN].trel = TFLOAT;
    xtP[CSC].unary = TRUE;  
    xtP[CSC].targ = TFLOAT; 
    xtP[CSC].trel = TFLOAT;
    xtP[SECANT].unary = TRUE;  
    xtP[SECANT].targ = TFLOAT; 
    xtP[SECANT].trel = TFLOAT;
    xtP[COT].unary = TRUE;  
    xtP[COT].targ = TFLOAT; 
    xtP[COT].trel = TFLOAT;
    xtP[ARCSIN].unary = TRUE;   
    xtP[ARCSIN].targ = TFLOAT;  
    xtP[ARCSIN].trel = TFLOAT;
    xtP[ARCCOS].unary = TRUE;   
    xtP[ARCCOS].targ = TFLOAT;  
    xtP[ARCCOS].trel = TFLOAT;
    xtP[ARCCSC].unary = TRUE;   
    xtP[ARCCSC].targ = TFLOAT;  
    xtP[ARCCSC].trel = TFLOAT;
    xtP[ARCSEC].unary = TRUE;   
    xtP[ARCSEC].targ = TFLOAT;  
    xtP[ARCSEC].trel = TFLOAT;
    xtP[ARCCOT].unary = TRUE;   
    xtP[ARCCOT].targ = TFLOAT;  
    xtP[ARCCOT].trel = TFLOAT;
    xtP[SQRT].unary = TRUE; 
    xtP[SQRT].targ = TFLOAT;    
    xtP[SQRT].trel = TFLOAT;
    xtP[ABS].unary = TRUE;  
    xtP[ABS].targ = TFLOAT; 
    xtP[ABS].trel = TFLOAT;
    xtP[EXP].unary = TRUE;  
    xtP[EXP].targ = TFLOAT; 
    xtP[EXP].trel = TFLOAT;
    xtP[LOG].unary = TRUE;  
    xtP[LOG].targ = TFLOAT; 
    xtP[LOG].trel = TFLOAT;
    xtP[ALOG].unary = TRUE; 
    xtP[ALOG].targ = TFLOAT;    
    xtP[ALOG].trel = TFLOAT;
    xtP[LN].unary = TRUE;   
    xtP[LN].targ = TFLOAT;  
    xtP[LN].trel = TFLOAT;
    xtP[SINH].unary = TRUE; 
    xtP[SINH].targ = TFLOAT;    
    xtP[SINH].trel = TFLOAT;        
    xtP[COSH].unary = TRUE; 
    xtP[COSH].targ = TFLOAT;    
    xtP[COSH].trel = TFLOAT;
    xtP[TANH].unary = TRUE; 
    xtP[TANH].targ = TFLOAT;    
    xtP[TANH].trel = TFLOAT;
    xtP[GAMMA].unary = TRUE;    
    xtP[GAMMA].targ = TFLOAT;   
    xtP[GAMMA].trel = TFLOAT;       
    xtP[FUNCT].targ = TFLOAT;       
    xtP[FUNCT].trel = TFLOAT;
    xtP[FUNCT].unary = TRUE;        
    xtP[EFUNCT].trinary = TRUE;
    xtP[EFUNCT].sgenv = 2;  
    xtP[EFUNCT].targ = TFLOAT;  
    xtP[EFUNCT].trel = TMARK;
    xtP[COMP].unary = TRUE;
    xtP[BITCNT].unary = TRUE;
    xtP[ZLENGTH].unary = TRUE;
    xtP[ZLENGTHM].unary = TRUE; 
    xtP[ZLENGTHM].targ = TMARK;
    xtP[ZLENGTHF].unary = TRUE; 
    xtP[ZLENGTHF].targ = TXTYPE;    
    xtP[ZLENGTHF].trel = TINT;
    xtP[ZLENGTHF].txop = TRUE;
    xtP[ZPOSITION].unary = TRUE;    
    xtP[ZPOSITION].targ = TMARK;
    xtP[ZNICONS].unary = TRUE;
    xtP[ZNICONS].targ = TMARK;
    xtP[ZCODE].unary = TRUE;    
    xtP[ZCODE].targ = TMARK;
    xtP[ZCHAR].unary = TRUE;    
    xtP[ZCHAR].trel = TMARK;
    xtP[ZFIRST].unary = TRUE;   
    xtP[ZFIRST].targ = TMARK;   
    xtP[ZFIRST].trel = TMARK;
    xtP[ZLAST].unary = TRUE;    
    xtP[ZLAST].targ = TMARK;    
    xtP[ZLAST].trel = TMARK;
    xtP[ZBASE].unary = TRUE;    
    xtP[ZBASE].targ = TMARK;    
    xtP[ZBASE].trel = TMARK;
    xtP[ZNEXT].unary = TRUE;    
    xtP[ZNEXT].targ = TMARK;    
    xtP[ZNEXT].trel = TMARK;
    xtP[ZPREVIOUS].unary = TRUE;
    xtP[ZPREVIOUS].targ = TMARK;    
    xtP[ZPREVIOUS].trel = TMARK;
    xtP[ZCOPY].unary = TRUE;    
    xtP[ZCOPY].targ = TMARK;    
    xtP[ZCOPY].trel = TMARK;
    xtP[ZALTERED].unary = TRUE; 
    xtP[ZALTERED].targ = TMARK;
    xtP[ZEXTENT].targ = TMARK;  
    xtP[ZEXTENT].trel = TMARK;
    xtP[ZSAMEMARK].targ = TMARK;
    xtP[ZPRECEDE].targ = TMARK;
    xtP[ZSEARCH].targ = TMARK;  
    xtP[ZSEARCH].trel = TMARK;
    xtP[ZSETMARK].trinary = TRUE;
    xtP[ZSETMARK].targ = TINT;  
    xtP[ZSETMARK].trel = TMARK;
    xtP[ZTEXTAT].trinary = TRUE;
    xtP[ZTEXTAT].targ = TINT;
    xtP[ZTEXTAT].trel = TMARK;
    xtP[ZHASSTYLE].targ = TMARK;  
    xtP[ZICONCODE].targ = TMARK;  
    xtP[ZICONFILE].targ = TMARK;
    xtP[ZICONFILE].trel = TMARK; 
    xtP[ZRGBN].trinary = TRUE;
    xtP[ZRGBN].targ = TFLOAT;   
    xtP[ZRGBN].trel = TINT;
    xtP[ZHSVN].trinary = TRUE;
    xtP[ZHSVN].targ = TFLOAT;   
    xtP[ZHSVN].trel = TINT;
    xtP[ZSTART].unary = TRUE;   
    xtP[ZSTART].targ = TMARK;   
    xtP[ZSTART].trel = TMARK;
    xtP[ZEND].unary = TRUE;     
    xtP[ZEND].targ = TMARK;     
    xtP[ZEND].trel = TMARK;
    xtP[ZKS].unary = TRUE;  
    xtP[ZKS].targ = TINT;   
    xtP[ZKS].trel = TMARK;
    xtP[ITOC].unary = TRUE;
    xtP[FTOC].targ = TFLOAT;        
    xtP[FTOC].unary = TRUE;
    xtP[STORINF].op = FALSE;
    xtP[STORINF].sgenv = 18;
    xtP[TKEYWORD].op = FALSE;
    xtP[TKEYWORD].sgenv = 2;
    xtP[TEXTARG].op = FALSE;
    xtP[TEXTARG].sgenv = -5;
    xtP[TEXTARG1].op = FALSE;
    xtP[TEXTARG1].sgenv = -6;
    xtP[ULOCADDR].op = FALSE;
    xtP[ULOCADDR].sgenv = 2;
    xtP[ULOCTABL].op = FALSE;
    xtP[ULOCTABL].sgenv = -4;
    xtP[COTOFN].unary = TRUE;
    xtP[SKIP].op = FALSE;
    xtP[TDOT].op = FALSE;
    xtP[TPOINT].op = FALSE;
    xtP[ZXYA].op = FALSE;
    xtP[ZXYR].op = FALSE;
    xtP[ZXYG].op = FALSE;
    xtP[ZXYI].op = FALSE;
    xtP[ZTEXTSEL].unary = TRUE; 
    xtP[ZTEXTSEL].targ = TXTYPE;    
    xtP[ZTEXTSEL].trel = TMARK;
    xtP[ZTEXTSEL].txop = TRUE;
    xtP[ZTEXTVIS].unary = TRUE; 
    xtP[ZTEXTVIS].targ = TXTYPE;    
    xtP[ZTEXTVIS].trel = TMARK;
    xtP[ZTEXTVIS].txop = TRUE;
    xtP[ZEDITBASE].unary = TRUE;    
    xtP[ZEDITBASE].targ = TXTYPE;   
    xtP[ZEDITBASE].trel = TMARK;
    xtP[ZEDITBASE].txop = TRUE;
    xtP[ZFILENAME].unary = TRUE;    
    xtP[ZFILENAME].targ = TXTYPE;   
    xtP[ZFILENAME].trel = TMARK;
    xtP[ZFILENAME].txop = TRUE;
    xtP[ZFILEPATH].unary = TRUE;    
    xtP[ZFILEPATH].targ = TXTYPE;   
    xtP[ZFILEPATH].trel = TMARK;
    xtP[ZFILEPATH].txop = TRUE;
    xtP[ZSWIDTH].unary = TRUE;
    xtP[ZSWIDTH].targ = TXTYPE;
    xtP[ZSWIDTH].trel = TINT;
    xtP[ZSWIDTH].txop = TRUE;
    xtP[ZSHEIGHT].unary = TRUE;
    xtP[ZSHEIGHT].targ = TXTYPE;
    xtP[ZSHEIGHT].trel = TINT;
    xtP[ZSHEIGHT].txop = TRUE;
    xtP[ZVALUEB].unary = TRUE;  
    xtP[ZVALUEB].targ = TXTYPE; 
    xtP[ZVALUEB].trel = TINT;
    xtP[ZVALUEB].txop = TRUE;
    xtP[ZVALUES].unary = TRUE;  
    xtP[ZVALUES].targ = TXTYPE; 
    xtP[ZVALUES].trel = TFLOAT;
    xtP[ZVALUES].txop = TRUE;
    xtP[ZVALUEDDE].unary = TRUE;
    xtP[ZVALUEDDE].targ = TXTYPE;
    xtP[ZVALUEDDE].trel = TMARK;
    xtP[ZVALUEDDE].txop = TRUE;
    xtP[CASETAB].op = FALSE;
    xtP[CASETAB].sgenv = -7;
    xtP[ZHOTINFOE].unary = TRUE; 
    xtP[ZHOTINFOE].targ = TXTYPE;    
    xtP[ZHOTINFOE].trel = TMARK;
    xtP[ZHOTINFOE].txop = TRUE;
    xtP[ZHOTINFOM].unary = TRUE; 
    xtP[ZHOTINFOM].targ = TMARK;    
    xtP[ZHOTINFOM].trel = TMARK;
    xtP[ZHOTINFOM].txop = TRUE;
    xtP[ZHOTSEL].unary = TRUE;  
    xtP[ZHOTSEL].targ = TXTYPE; 
    xtP[ZHOTSEL].trel = TMARK;
    xtP[ZHOTSEL].txop = TRUE;
    xtP[ZNEXTWORD].unary = TRUE;    
    xtP[ZNEXTWORD].targ = TMARK;    
    xtP[ZNEXTWORD].trel = TMARK;
    xtP[ZNEXTLINE].unary = TRUE;    
    xtP[ZNEXTLINE].targ = TMARK;    
    xtP[ZNEXTLINE].trel = TMARK;
    xtP[ZNEXTNUM].unary = TRUE;    
    xtP[ZNEXTNUM].targ = TMARK;    
    xtP[ZNEXTNUM].trel = TFLOAT;
    xtP[ZNEXTEXPR].unary = TRUE;    
    xtP[ZNEXTEXPR].targ = TMARK;    
    xtP[ZNEXTEXPR].trel = TMARK;
	xtP[ZFACTORIAL].unary = TRUE;
	xtP[ZFACTORIAL].targ = TINT;
	xtP[ZFACTORIAL].trel = TFLOAT;
	xtP[ZCOMB].targ = TINT;
	xtP[ZCOMB].trel = TFLOAT;
	
    /* initialize command error table */
    
    for(i=0; i<C_MAX; i++) 
        cmd_err_tab[i] = FALSE;
    
    cmd_err_tab[MUNIT] = TRUE;
    cmd_err_tab[MOD] = TRUE;
    cmd_err_tab[ARCSIN] = TRUE;
    cmd_err_tab[ARCCOS] = TRUE;
    cmd_err_tab[SQRT] = TRUE;
    cmd_err_tab[LOG] = TRUE;
    cmd_err_tab[LN] = TRUE;
    cmd_err_tab[ZCODE] = TRUE;
    cmd_err_tab[ZEXTENT] = TRUE;
    cmd_err_tab[ZSAMEMARK] = TRUE;
    cmd_err_tab[ZSETMARK] = TRUE;
    cmd_err_tab[ZPRECEDE] = TRUE;
    cmd_err_tab[EXPONENT] = TRUE;
    cmd_err_tab[GARRAYVAL] = TRUE;
    cmd_err_tab[GARRAYADDR] = TRUE;
    cmd_err_tab[IGARRAYVAL] = TRUE;
    cmd_err_tab[IGARRAYADDR] = TRUE;
    cmd_err_tab[BGARRAYVAL] = TRUE;
    cmd_err_tab[BGARRAYADDR] = TRUE;
    cmd_err_tab[MGARRAYVAL] = TRUE;
    cmd_err_tab[MGARRAYADDR] = TRUE;
    cmd_err_tab[XGARRAYVAL] = TRUE;
    cmd_err_tab[LARRAYVAL] = TRUE;
    cmd_err_tab[LARRAYADDR] = TRUE;
    cmd_err_tab[ILARRAYVAL] = TRUE;
    cmd_err_tab[ILARRAYADDR] = TRUE;
    cmd_err_tab[BLARRAYVAL] = TRUE;
    cmd_err_tab[BLARRAYADDR] = TRUE;
    cmd_err_tab[MLARRAYVAL] = TRUE;
    cmd_err_tab[MLARRAYADDR] = TRUE;
    cmd_err_tab[XLARRAYVAL] = TRUE;
    cmd_err_tab[PARRAYVAL] = TRUE;
    cmd_err_tab[PARRAYADDR] = TRUE;
    cmd_err_tab[IPARRAYVAL] = TRUE;
    cmd_err_tab[IPARRAYADDR] = TRUE;
    cmd_err_tab[BPARRAYVAL] = TRUE;
    cmd_err_tab[BPARRAYADDR] = TRUE;
    cmd_err_tab[MPARRAYVAL] = TRUE;
    cmd_err_tab[MPARRAYADDR] = TRUE;
    cmd_err_tab[XPARRAYVAL] = TRUE;
    cmd_err_tab[GDYARRAY] = TRUE;
    cmd_err_tab[IGDYARRAY] = TRUE;
    cmd_err_tab[BGDYARRAY] = TRUE;
    cmd_err_tab[LDYARRAY] = TRUE;
    cmd_err_tab[ILDYARRAY] = TRUE;
    cmd_err_tab[BLDYARRAY] = TRUE;
    cmd_err_tab[PDYARRAY] = TRUE;
    cmd_err_tab[IPDYARRAY] = TRUE;
    cmd_err_tab[BPDYARRAY] = TRUE;
    cmd_err_tab[GDYARRAYVAL] = TRUE;
    cmd_err_tab[IGDYARRAYVAL] = TRUE;
    cmd_err_tab[BGDYARRAYVAL] = TRUE;
    cmd_err_tab[GDYARRAYADDR] = TRUE;
    cmd_err_tab[IGDYARRAYADDR] = TRUE;
    cmd_err_tab[BGDYARRAYADDR] = TRUE;
    cmd_err_tab[LDYARRAYVAL] = TRUE;
    cmd_err_tab[ILDYARRAYVAL] = TRUE;
    cmd_err_tab[BLDYARRAYVAL] = TRUE;
    cmd_err_tab[LDYARRAYADDR] = TRUE;
    cmd_err_tab[ILDYARRAYADDR] = TRUE;
    cmd_err_tab[BLDYARRAYADDR] = TRUE;
    cmd_err_tab[PDYARRAYVAL] = TRUE;
    cmd_err_tab[IPDYARRAYVAL] = TRUE;
    cmd_err_tab[BPDYARRAYVAL] = TRUE;
    cmd_err_tab[PDYARRAYADDR] = TRUE;
    cmd_err_tab[IPDYARRAYADDR] = TRUE;
    cmd_err_tab[BPDYARRAYADDR] = TRUE;
    cmd_err_tab[ZICONCODE] = TRUE;
    cmd_err_tab[ZICONFILE] = TRUE;
    cmd_err_tab[ZFACTORIAL] = TRUE;
    cmd_err_tab[ZCOMB] = TRUE;
    cmd_err_tab[C_FONT] = TRUE;
    cmd_err_tab[C_ICONS] = TRUE;
    cmd_err_tab[C_PATTERN] = TRUE;
    cmd_err_tab[C_CURSOR] = TRUE;
    cmd_err_tab[C_RGB] = TRUE;
    cmd_err_tab[C_HSV] = TRUE;
    cmd_err_tab[C_WCOLOR] = TRUE;
    cmd_err_tab[C_COLOR] = TRUE;
    cmd_err_tab[C_BOUNDS] = TRUE;
    cmd_err_tab[C_MOVE] = TRUE;
    cmd_err_tab[C_RMOVE] = TRUE;
    cmd_err_tab[C_GMOVE] = TRUE;
    cmd_err_tab[C_PRESS] = TRUE;
    cmd_err_tab[C_RANDU] = TRUE;
    cmd_err_tab[C_WRITE1] = TRUE;
    cmd_err_tab[C_WRITE2] = TRUE;
    cmd_err_tab[C_SHOW] = TRUE;
    cmd_err_tab[C_SHOWB] = TRUE;
    cmd_err_tab[C_SHOWO] = TRUE;
    cmd_err_tab[C_SHOWH] = TRUE;
    cmd_err_tab[C_SHOWT] = TRUE;
    cmd_err_tab[C_RTEXT] = TRUE;
    cmd_err_tab[C_RCIRCLE] = TRUE;
    cmd_err_tab[C_RCIRCLEB] = TRUE;
    cmd_err_tab[C_GTEXT] = TRUE;
    cmd_err_tab[C_SCALEX] = TRUE;
    cmd_err_tab[C_SCALEY] = TRUE;
    cmd_err_tab[C_LABELX] = TRUE;
    cmd_err_tab[C_LABELY] = TRUE;
    cmd_err_tab[C_AXES] = TRUE;
    cmd_err_tab[C_MENU] = TRUE;
    cmd_err_tab[C_VSET] = TRUE;
    cmd_err_tab[C_SETFILE] = TRUE;
    cmd_err_tab[C_ADDFILE] = TRUE;
    cmd_err_tab[C_DATAIN] = TRUE;
    cmd_err_tab[C_DATAOUT] = TRUE;
    cmd_err_tab[C_XIN] = TRUE;
    cmd_err_tab[C_XOUT] = TRUE;
    cmd_err_tab[C_SOUND] = TRUE;
    cmd_err_tab[C_BLOCK] = TRUE;
    cmd_err_tab[C_PICT] = TRUE;
    cmd_err_tab[C_ZERO] = TRUE;
    cmd_err_tab[C_ZEROM] = TRUE;
    cmd_err_tab[C_SET] = TRUE;
    cmd_err_tab[C_SETC] = TRUE;
    cmd_err_tab[C_EXEC] = TRUE;
    cmd_err_tab[C_DO] = TRUE;
    cmd_err_tab[C_JUMP] = TRUE;
    cmd_err_tab[C_PREANS] = TRUE;
    cmd_err_tab[C_JUDGE] = TRUE;
    cmd_err_tab[C_RESET] = TRUE;
    cmd_err_tab[C_ENDUNIT] = TRUE;
    cmd_err_tab[C_NEWPAL] = TRUE;
    cmd_err_tab[C_PALETTE] = TRUE;
    cmd_err_tab[C_SLIDER] = TRUE;
    cmd_err_tab[C_RSLIDER] = TRUE;
    cmd_err_tab[C_GSLIDER] = TRUE;
    cmd_err_tab[C_RBUTTON] = TRUE;
    cmd_err_tab[C_GBUTTON] = TRUE;
    cmd_err_tab[C_SLIDERR] = TRUE;
    cmd_err_tab[C_RSLIDERR] = TRUE;
    cmd_err_tab[C_GSLIDERR] = TRUE;
    cmd_err_tab[C_RBUTTONR] = TRUE;
    cmd_err_tab[C_GBUTTONR] = TRUE;
    cmd_err_tab[C_EDIT] = TRUE;
    cmd_err_tab[C_REDIT] = TRUE;
    cmd_err_tab[C_GEDIT] = TRUE;
    cmd_err_tab[C_EDITR] = TRUE;
    cmd_err_tab[C_REDITR] = TRUE;
    cmd_err_tab[C_GEDITR] = TRUE;
    cmd_err_tab[C_ALLOC] = TRUE;
    cmd_err_tab[C_GET] = TRUE;
    cmd_err_tab[C_RGET] = TRUE;
    cmd_err_tab[C_GGET] = TRUE;
    cmd_err_tab[C_PUT] = TRUE;
    cmd_err_tab[C_RPUT] = TRUE;
    cmd_err_tab[C_GPUT] = TRUE;
	cmd_err_tab[C_STYLE] = TRUE;
	cmd_err_tab[C_VIDEO] = TRUE;
    
#ifdef MAC
    icgen68K(); /* initialize for 68000 code generator */
	icgenPPC(); /* initialize for PPC   code generator */
#endif

} /* icompile */

/* ******************************************************************* */
